package ap;

import ap.parameters.{GoalSettings, PreprocessingSettings, Param}
import ap.parser.{InputAbsy2Internal, Parser2InputAbsy, Preprocessing}
import ap.terfor.{Formula, TermOrder}
import ap.terfor.conjunctions.{Conjunction, Quantifier}
import ap.proof.{ModelSearchProver, ExhaustiveProver, ConstraintSimplifier,
                 Timeout}
import ap.proof.tree.ProofTree
import ap.proof.goal.{Goal, SymbolWeights}

abstract class AbstractFileProver(reader : java.io.Reader, output : Boolean,
                                  timeout : Int, userDefStoppingCond : => Boolean,
                                  preprocSettings : PreprocessingSettings,
                                  initialGoalSettings : GoalSettings) {

  protected def println(str : => String) : Unit = (if (output) Predef.println(str))
  
  val (inputFormula, eliminatedConstants, order) = {
    val (f, consts, order) = Parser2InputAbsy(reader)
    reader.close
    println("Preprocessing ...")
    Preprocessing(f, consts, order, preprocSettings)
  }
  
  val formula = InputAbsy2Internal(inputFormula, order)

  private val goalSettings =
    Param.SYMBOL_WEIGHTS.set(initialGoalSettings,
                             SymbolWeights.normSymbolFrequencies(formula, 1000))
  
  protected def findModelTimeout : Conjunction = {
    println("Constructing satisfying assignment for the existential constants ...")
    findCounterModelTimeout(Conjunction.negate(formula, order))
  }
  
  protected def findCounterModelTimeout : Conjunction = {
    println("Constructing countermodel ...")
    findCounterModelTimeout(formula)
  }
  
  private def findCounterModelTimeout(f : Formula) = {
    val timeBefore = System.currentTimeMillis
    val stoppingCond = () => {
      if ((System.currentTimeMillis - timeBefore > timeout) || userDefStoppingCond)
        Timeout.raise
    }

    Timeout.withChecker(stoppingCond) {
      ModelSearchProver(f, order, goalSettings)
    }
  }
  
  protected def findModel(f : Conjunction) : Conjunction =
    ModelSearchProver(f.negate, f.order)
  
  protected def constructProofTree(mostGeneralConstraint : Boolean) : ProofTree = {
    // explicitly quantify all universal variables
    
    val closedFor =
      Conjunction.quantify(Quantifier.ALL, eliminatedConstants, formula, order)
    
    println("Proving ...")
    
    val timeBefore = System.currentTimeMillis
    val stoppingCond = () => {
      if ((System.currentTimeMillis - timeBefore > timeout) ||
          userDefStoppingCond)
        Timeout.raise
    }

    Timeout.withChecker(stoppingCond) {
      new ExhaustiveProver(!mostGeneralConstraint, goalSettings)(closedFor, order)
    }
  }
}
