package ap.proof;

import ap.basetypes.IdealInt
import ap.terfor.{Formula, TermOrder, ConstantTerm}
import ap.terfor.conjunctions.{Conjunction, Quantifier}
import ap.terfor.linearcombination.LinearCombination
import ap.terfor.equations.EquationConj
import ap.terfor.substitutions.{Substitution, IdentitySubst}
import ap.proof.goal.{Goal, FactsNormalisationTask, EliminateFactsTask,
                      AddFactsTask, UpdateTasksTask}
import ap.util.{Debug, Logic, LRUCache, FilterIt, Seqs}
import ap.parameters.{GoalSettings, Param}
import ap.proof.tree._

import scala.collection.mutable.Stack


/**
 * A prover that tries to construct a countermodel of a ground formula. This
 * prover works in depth-first mode, in contrast to the
 * <code>ExhaustiveProver</code>.
 */
object ModelSearchProver {

  private val AC = Debug.AC_PROVER
   
  private val simplifier = ConstraintSimplifier.FAIR_SIMPLIFIER
  
  // we need to store eliminated facts from goals, otherwise we could not
  // construct a complete model
  private val ptf = new SimpleProofTreeFactory(true, simplifier) {
    override def eliminatedConstant(subtree : ProofTree,
                                    c : ConstantTerm,
                                    witness : (Substitution, TermOrder) => Substitution,
                                    vocabulary : Vocabulary) : ProofTree =
      new WitnessTree (subtree, c, witness, vocabulary)
  }

  private val cache = new LRUCache[Formula, Conjunction] (1000)
  
  /**
   * <code>inputFor</code> is the formula to be disproven. The result of the
   * method is a countermodel of <code>inputFor</code>, or <code>FALSE</code>
   * if it was not possible to find one (this implies that <code>inputFor</code>
   * is valid).
   */
  def apply(inputFor : Formula, order : TermOrder) : Conjunction = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(ModelSearchProver.AC,
                    inputFor.variables.isEmpty && (order isSortingOf inputFor))
    ////////////////////////////////////////////////////////////////////////////
    cache(inputFor) {
      applyHelp(inputFor, order, GoalSettings.DEFAULT)
    }
  }

  def apply(inputFor : Formula, order : TermOrder,
            settings : GoalSettings) : Conjunction = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(ModelSearchProver.AC,
                    inputFor.variables.isEmpty && (order isSortingOf inputFor))
    ////////////////////////////////////////////////////////////////////////////
    applyHelp(inputFor, order,
              Param.CONSTRAINT_SIMPLIFIER.set(settings, simplifier))
  }
   
  /**
   * <code>inputFor</code> is the formula to be disproven. The result of the
   * method is a countermodel of <code>inputFor</code>, or <code>None</code>
   * if it was not possible to find one (this implies that <code>inputFor</code>
   * is valid).
   */
  private def applyHelp(inputFor : Formula, order : TermOrder,
                        settings : GoalSettings) : Conjunction = {
    val goal = Goal(inputFor, order.orderedConstants, order, settings)
//    val model = findModelFair(goal, 500)
    val model = findModel(goal, List(), Set(), 0)
    if (model == null)
      Conjunction.FALSE
    else
      model
  }

  private def isSatisfiableGoal(tree : ProofTree) : Boolean = tree match {
    case goal : Goal => {
      val c = goal.closingConstraint
      if (c.isTrue) {
        true
      } else if (c.isFalse) {
        false
      } else {
        assert(false) // not supposed to happen
        false
      }
    }
    case _ => false
  }
  
  //////////////////////////////////////////////////////////////////////////////
  //
  // Model search in a fair manner, which means that the search in not purely
  // depth-first, but after a certain amount of time the prover switches to a
  // different branch. This is currently not used, but might sometimes by more
  // robust than the pure depth-first search
  
  private case class GoalSet(goal : Goal,
                             witnesses : List[(Substitution, TermOrder) => Substitution],
                             constsToIgnore : Set[ConstantTerm])
  
  private def findModelFair(initialGoal : Goal, stepsPerGoal : Int) : Conjunction = {
    val goals = new scala.collection.mutable.Queue[GoalSet]
    goals += GoalSet(initialGoal, List(), Set())

    while (!goals.isEmpty) {
      Timeout.check

      var GoalSet(goal, witnesses, constsToIgnore) = goals.dequeue
      
      val localGoals = new scala.collection.mutable.Stack[GoalSet]

      def destructTree(tree : ProofTree,
                       witnesses : List[(Substitution, TermOrder) => Substitution],
                       constsToIgnore : Set[ConstantTerm]) : Unit = tree match {
        case goal : Goal =>
          localGoals += GoalSet(goal, witnesses, constsToIgnore)

        case tree : WitnessTree =>
          destructTree(tree.subtree, tree.witness :: witnesses, constsToIgnore)

        case tree : QuantifiedTree => {
          //////////////////////////////////////////////////////////////////////
          Debug.assertInt(ModelSearchProver.AC, tree.quan == Quantifier.ALL)
          //////////////////////////////////////////////////////////////////////
          destructTree(tree.subtree,
                       witnesses,
                       constsToIgnore ++ tree.quantifiedConstants)
        }

        case _ : WeakenTree =>
          assert(false)  // should not happen

        case tree : ProofTreeOneChild => 
          destructTree(tree.subtree, witnesses, constsToIgnore)
     
        case tree : AndTree => {
          destructTree(tree.right, witnesses, constsToIgnore)
          destructTree(tree.left, witnesses, constsToIgnore)
        }
      }

      var i : Int = 0
      while (goal != null && i < stepsPerGoal) {
        if (isSatisfiableGoal(goal)) {
          // nothing to do here ...
          goal = null
        } else if (goal.stepPossible) {
          goal.step(ptf) match {
            case newGoal : Goal =>
              goal = newGoal
            case tree : ProofTree => {
              destructTree(tree, witnesses, constsToIgnore)
              goal = null
            }
          }
        } else {
          return constructModel(goal.facts,
                                witnesses, constsToIgnore,
                                goal.order)
        }
        
        if (goal == null && !localGoals.isEmpty) {
          val GoalSet(newGoal, newWit, newConsts) = localGoals.pop
          goal = newGoal
          witnesses = newWit
          constsToIgnore = newConsts
        }
        
        i = i + 1
      }
      
      if (goal != null)
        goals += GoalSet(goal, witnesses, constsToIgnore)
      goals ++= localGoals
    }
    
    null
  }
  
  //////////////////////////////////////////////////////////////////////////////
  
  private def findModel(tree : ProofTree,
                        // functions to reconstruct witnesses for eliminated
                        // constants
                        witnesses : List[(Substitution, TermOrder) => Substitution],
                        // explicitly quantified constants that do not have to
                        // be included in models
                        constsToIgnore : Set[ConstantTerm],
                        depth : Int) : Conjunction = {
    Timeout.check
    
    if (isSatisfiableGoal(tree)) {
//      println("backtracking (depth " + depth + ")")
      null
    } else
    tree match { 
      case goal : Goal =>
        if (goal.stepPossible)
          findModel(goal.step(ptf), witnesses, constsToIgnore, depth)
        else
          constructModel(goal.facts,
                         witnesses, constsToIgnore,
                         goal.order)
        
      case tree : WitnessTree =>
        findModel(tree.subtree, tree.witness :: witnesses, constsToIgnore, depth)

      case tree : ProofTreeOneChild => {
        ////////////////////////////////////////////////////////////////////////
        Debug.assertInt(ModelSearchProver.AC, tree match {
                          case _ : WeakenTree => false
                          case tree : QuantifiedTree => tree.quan == Quantifier.ALL
                          case _ => true
                        })
        ////////////////////////////////////////////////////////////////////////
        val newConstsToIgnore = tree match {
          case tree : QuantifiedTree => constsToIgnore ++ tree.quantifiedConstants
          case _ => constsToIgnore
        }
        findModel(tree.subtree, witnesses, newConstsToIgnore, depth)
      }
     
      case tree : AndTree => {
        val model = findModel(tree.left, witnesses, constsToIgnore, depth + 1)
        if  (model == null)
          findModel(tree.right, witnesses, constsToIgnore, depth + 1)
        else
          model
      }
    }
  }

  private def constructModel(facts : Conjunction,
                             witnesses : List[(Substitution, TermOrder) => Substitution],
                             constsToIgnore : Set[ConstantTerm],
                             order : TermOrder) : Conjunction = {
    val constantValues : Substitution =
      (new IdentitySubst(order).asInstanceOf[Substitution] /: witnesses)(
                                                         (s, w) => w(s, order))
    
    val valuesConj =
      EquationConj(for (c <- (order.orderedConstants -- constsToIgnore).elements)
                   yield LinearCombination(Array((IdealInt.ONE, c),
                                                 (IdealInt.MINUS_ONE, constantValues(c))),
                                           order),
                   order)
    Conjunction.conj(Array(valuesConj, facts), order)
  }
}

 
private case class WitnessTree(val subtree : ProofTree,
                               val eliminatedConstant : ConstantTerm,
                               val witness : (Substitution, TermOrder) => Substitution,
                               val vocabulary : Vocabulary)
                   extends { protected val subtreeOrder = vocabulary.order }
                           with ProofTreeOneChild {

  def update(newSubtree : ProofTree, newConstantFreedom : ConstantFreedom) : ProofTree =
    new WitnessTree(subtree, eliminatedConstant, witness,
                    vocabulary updateConstantFreedom newConstantFreedom)

  lazy val closingConstraint : Conjunction =
    subtree.closingConstraint

  lazy val closingConstantFreedom : ConstantFreedom =
    subtree.closingConstantFreedom

  lazy val fixedConstantFreedom : Boolean =
    subtree.fixedConstantFreedom

  lazy val stepMeaningful : Boolean =
    subtree.stepMeaningful
  
  def newConstantFreedomForSubtree(cf : ConstantFreedom) : ConstantFreedom = cf
}