package ap.proof

import scala.util.DynamicVariable

object Timeout {
  
  private val checker : DynamicVariable[() => Unit] = new DynamicVariable(() => {})
  
  def raise : Unit = throw Timeout(None)
  
  /**
   * Check whether a timeout has occurred. In this case, on object of class
   * <code>Timeout</code> is thrown
   */
  def check : Unit = checker.value.apply

  /**
   * Execute code in the context of a given timeout checker <code>c</code>,
   * which is a function that raises <code>Timeout</code> in case a timeout
   * occurred
   */
  def withChecker[A](c : () => Unit)(comp : => A) : A =
    checker.withValue(c)(comp)
  
  /**
   * Apply a conversion function to the argument of a possible raised timeout.
   * The idea is that the timeout argument should describe the partial
   * (unfinished) results that had been computed before the timeout
   */
  def unfinished[A](comp : => A)(errorHandler : (Any) => Any) : A =
    try { comp } catch {
      case Timeout(x) => {
        // avoid nested timeouts
        val newRes = withChecker(() => {}) (errorHandler(x))
        throw Timeout(newRes)
      }
    }

  def unfinishedValue[A](errorValue : Any)(comp : => A) : A =
    try { comp } catch {
      case Timeout(_) => throw Timeout(errorValue)
    }
  
  def catchTimeout[A](comp : => A)(errorHandler : (Any) => A) : A =
    try { comp } catch {
      case Timeout(x) => errorHandler(x)
    }
  
}

/**
 * Object that is thrown in case of a timeout (or the user stopped the proof
 * search)
 */
case class Timeout(unfinishedResult : Any) extends Exception
