package ap.proof.goal;

import scala.collection.mutable.ArrayBuffer

import ap.terfor.{Formula, ConstantTerm, VariableTerm,
                  TermOrder, TerFor, Term}
import ap.terfor.conjunctions.{Conjunction, NegatedConjunctions,
                               ReduceWithConjunction, Quantifier}
import ap.terfor.arithconj.ArithConj
import ap.terfor.equations.{EquationConj, NegEquationConj}
import ap.terfor.inequalities.InEqConj
import ap.terfor.preds.{PredConj, Atom}
import ap.terfor.linearcombination.LinearCombination
import ap.terfor.substitutions.{Substitution, IdentitySubst}
import ap.util.{Debug, FilterIt, Logic, Seqs}
import ap.parameters.{GoalSettings, Param}
import ap.proof.tree.{ProofTree, ProofTreeFactory}

object Goal {
  
  private val AC = Debug.AC_GOAL

  def apply(facts : Conjunction, compoundFormulas : CompoundFormulas,
            tasks : TaskManager, age : Int,
            eliminatedConstants : Set[ConstantTerm], vocabulary : Vocabulary,
            definedSyms : Substitution, settings : GoalSettings) : Goal =
    if (facts.isFalse)
      TRUE(vocabulary)
    else
      new Goal(facts, compoundFormulas, tasks, age, eliminatedConstants,
               vocabulary, definedSyms, settings)

  def apply(f : Formula, order : TermOrder, settings : GoalSettings) : Goal =
    apply(f, Set.empty, order, settings)
  
  def apply(f : Formula,
            eliminatedConstants : Set[ConstantTerm],
            order : TermOrder,
            settings : GoalSettings) : Goal = {
    // TODO: is this the right place for reduction?
    val reducedF =
      ReduceWithConjunction(Conjunction.TRUE, order)(Conjunction.conj(f, order))
    val tasks = formulaTasks(reducedF, 0, eliminatedConstants, Vocabulary(order), settings)
    apply(Conjunction.TRUE,
          CompoundFormulas.EMPTY,
          TaskManager.EMPTY ++ tasks,
          0,
          eliminatedConstants,
          Vocabulary(order),
          new IdentitySubst (order),
          settings)
  }
  
  def TRUE(vocabulary : Vocabulary) : Goal =
    new Goal (Conjunction.FALSE,
              CompoundFormulas.EMPTY,
              TaskManager.EMPTY, 0,
              Set.empty, vocabulary,
              new IdentitySubst (vocabulary.order),
              GoalSettings.DEFAULT)
  
  val TRUE : Goal = TRUE(Vocabulary.EMPTY)
  
  /**
   * Create the tasks to store and handle an arbitrary given formula. This
   * method is part of the goal because different tasks might be created
   * depending on the settings
   */
  def formulaTasks(formula : Conjunction,
                   age : Int,
                   eliminatedConstants : Set[ConstantTerm],
                   vocabulary : Vocabulary,
                   settings : GoalSettings) : Seq[FormulaTask] =
    if (formula.isFalse) {
      List()
    } else if (formula.isTrue || formula.isLiteral) {
      List(new AddFactsTask(formula, age))
    } else if (formula.isNegatedConjunction) {
      val disj = formula.negatedConjs(0)
      //////////////////////////////////////////////////////////////////////////
      // possibly existing quantifiers should have been pulled out and
      // should not occur at this point
      Debug.assertInt(Goal.AC, disj.quans.isEmpty)
      //////////////////////////////////////////////////////////////////////////

      (for (f <- disj.negatedConjs;
            g <- formulaTasks(f, age, eliminatedConstants, vocabulary, settings))
       yield g) ++
      (if (disj.arithConj.isTrue && disj.predConj.isTrue) {
         List()
       } else {
         val facts = Conjunction(List(), disj.arithConj, disj.predConj,
                                 NegatedConjunctions.TRUE, formula.order).negate
         List(new AddFactsTask(facts, age))})
    } else if (formula.quans.isEmpty) {
      List(BetaFormulaTask(formula, age, eliminatedConstants, vocabulary, settings))
    } else formula.quans.last match {
      case Quantifier.ALL => List(new AllQuantifierTask(formula, age))
      case Quantifier.EX =>
	if (formula.isDivisibility)
	  List(new DivisibilityTask(formula, age))
        else if (Param.POS_UNIT_RESOLUTION(settings) &&
                 (NegLitClauseTask isCoveredFormula formula))
	  List(new NegLitClauseTask(formula, age))
	else
	  List(new ExQuantifierTask(formula, age))
    }

}

class Goal private (val facts : Conjunction,
                    val compoundFormulas : CompoundFormulas,
                    val tasks : TaskManager,
                    val age : Int,
                    val eliminatedConstants : Set[ConstantTerm],
                    val vocabulary : Vocabulary,
                    val definedSyms : Substitution,
                    val settings : GoalSettings) extends ProofTree {

  //////////////////////////////////////////////////////////////////////////////
  Debug.assertCtor(Goal.AC,
                   // facts are really just literals (equations, inequalities,
                   // predicate literals)
                   facts.quans.isEmpty &&
                   facts.negatedConjs.isEmpty &&
                   // clauses are handled similar to facts; there are two
                   // kinds: quantified formulae that are instantiated
                   // through unit resolution with facts, and ground formulae
                   // that do not contain predicates or eliminated constants
                   // (splitting is unnecessary for such formulae)
                   compoundFormulas.qfClauses.forall((c) =>
                       constantFreedom.isShielded(c, bindingContext) ||
                       Seqs.disjoint(eliminatedConstants, c.constants) &&
                       Conjunction.collectQuantifiers(c).isEmpty) &&
                   // there should not be any free variables (use constants
                   // instead)
                   facts.variables.isEmpty &&
                   // if a contradiction is detected, we can forget about
                   // everything
                   (!facts.isFalse || tasks.isEmpty && compoundFormulas.isEmpty) &&
                   // if clauses need to be rematched, a task for this should
                   // have been generated
                   (!compoundFormulas.lazyQuantifiedClauses.factsAreOutdated(facts.predConj) ||
                     tasks.taskInfos.containsLazyMatchTask ||
                     {
                       // alternatively, it might happen that the next rule
                       // application closes the goal
                       val nextTask = tasks.max
                       nextTask.isInstanceOf[AddFactsTask] &&
                         nextTask.asInstanceOf[AddFactsTask].formula == Conjunction.TRUE
                     }) &&
                   // we must choose an ordering such that the constants to be
                   // eliminated are maximal
                   (facts isSortedBy order) &&
                   (compoundFormulas isSortedBy order) &&
                   (definedSyms isSortedBy order) &&
                   (order constantsAreMaximal eliminatedConstants))
  //////////////////////////////////////////////////////////////////////////////

  private def elimConstants(tfs : Seq[LinearCombination]) : Seq[LinearCombination] = {
    val res = new ArrayBuffer[LinearCombination]
    for (tf <- tfs) {
      if (!tf.isEmpty && !(this eliminates tf.leadingTerm)) {
        ////////////////////////////////////////////////////////////////////////
        Debug.assertInt(Goal.AC, Seqs.disjoint(tf.constants, eliminatedConstants))
        ////////////////////////////////////////////////////////////////////////
        res += tf
      }
    }
    res
  }
  
  private def elimConstants(tfs : Iterator[LinearCombination])
                                                : Iterator[LinearCombination] =
    FilterIt(tfs, (tf:LinearCombination) => {
      val res = tf.isEmpty || !(this eliminates tf.leadingTerm)
      //////////////////////////////////////////////////////////////////////////
      Debug.assertPost(Goal.AC,
                       res == Seqs.disjoint(tf.constants, eliminatedConstants))
      //////////////////////////////////////////////////////////////////////////
      res
    })
    
  private def arithClauses : NegatedConjunctions = compoundFormulas.qfClauses
    
  lazy val closingConstraint : Conjunction = {
    val ac = facts.arithConj
    
    // we can always leave out antecedent equations that contain eliminated
    // constants (because such equations can exhaustively be applied to all
    // other formulae)
    val selectedPosEqs =
      ac.positiveEqs.updateEqsSubset(elimConstants(ac.positiveEqs), order)
    
    if (Seqs.disjoint(eliminatedConstants, facts.predConj.constants,
                      ac.negativeEqs.constants) &&
        Seqs.disjoint(eliminatedConstants, facts.predConj.constants,
                      ac.inEqs.constants)) {
      // then also succedent equations and inequalities that contain eliminated
      // constants can be left out
          
      val selectedNegEqs =
        ac.negativeEqs.updateEqsSubset(elimConstants(ac.negativeEqs), order)
      val selectedInEqs =
        InEqConj(elimConstants(ac.inEqs.elements ++
                               ac.inEqs.geqZeroInfs.elements),
                 order)
      val selectedAC =
        ArithConj(selectedPosEqs, selectedNegEqs, selectedInEqs, order)
      val selectedConj =
        definedSyms(Conjunction(List(), selectedAC, PredConj.TRUE, arithClauses,
                                order)).negate        
      constantFreedom.unshieldedPart(selectedConj, bindingContext)
    } else {
      // otherwise, we have to consider all arithmetic literals, because
      // rules like OmegaElim or SplitEq are not applicable in the presence
      // of predicate literals, and the calculus is thus not complete for
      // eliminating constants
        
      val selectedAC = ac.updatePositiveEqs(selectedPosEqs, order)
      val selectedConj =
        definedSyms(Conjunction(List(), selectedAC, PredConj.TRUE, arithClauses,
                                order)).negate
      val reducedSelectedConj =
        constantFreedom.unshieldedPart(selectedConj, bindingContext)

      if (Seqs.disjoint(reducedSelectedConj.constants, eliminatedConstants)) {
        reducedSelectedConj
      } else {
        // we directly eliminate the constants from the PA formulae to avoid
        // violations of the constant freedom conditions
        val quantifiedConj =
          Conjunction.quantify(Quantifier.ALL,
                               order sort eliminatedConstants,
                               selectedConj, order)
        Param.CONSTRAINT_SIMPLIFIER(settings)(quantifiedConj, order)
      }
    }
  }
  
  lazy val closingConstantFreedom : ConstantFreedom =
    if (constantFreedom.isBottomWRT(constants -- eliminatedConstants))
      // if only eliminated constants are free constants, it is not possible
      // that the constant freedom changes
      constantFreedom
   else
      constantFreedom.findNonFreeness(closingConstraint, bindingContext)

  lazy val fixedConstantFreedom : Boolean =
    constantFreedom == closingConstantFreedom

  lazy val mayAlias : ((LinearCombination, LinearCombination) => Boolean) =
    new AliasAnalyser (reduceWithFacts,
                       constantFreedom, vocabulary.bindingContext,
                       order)
  
  //////////////////////////////////////////////////////////////////////////////

  /**
   * All constants that occur in formulas in this goal
   */
  lazy val constants : Set[ConstantTerm] =
    facts.constants ++ compoundFormulas.constants ++ tasks.taskInfos.constants
  
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Return whether <code>t</code> is a constant that is eliminated in this goal
   */
  def eliminates(t : Term) : Boolean = t match {
    case t : ConstantTerm => eliminatedConstants contains t
    case _ => false
  }

  lazy val reduceWithFacts : ReduceWithConjunction =
    ReduceWithConjunction(facts, order)
   
  //////////////////////////////////////////////////////////////////////////////

  val stepPossible : Boolean = !tasks.isEmpty

  val stepMeaningful : Boolean = stepPossible
  
  def step(ptf : ProofTreeFactory) : ProofTree = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(Goal.AC, this.stepPossible)
    ////////////////////////////////////////////////////////////////////////////
    
    // apply the first task to the goal and then stop immediately
    val task = tasks.max
//    println(task)
//    ap.util.Timer.measure(task.getClass.getName) {
      task(this, ptf)
//    }
  }

  /**
   * Create the tasks to store and handle an arbitrary given formula. This
   * method is part of the goal because different tasks might be created
   * depending on the settings
   */
  def formulaTasks(formula : Conjunction) : Seq[FormulaTask] =
    Goal.formulaTasks(formula, age, eliminatedConstants, vocabulary, settings)
      
  def updateConstantFreedom(cf : ConstantFreedom) : Goal =
    if (constantFreedom == cf) {
      this
    } else if (facts.isFalse) {
      // then it is meaningless to generate a new task, instead we just
      // change the vocabulary of the goal
      Goal.TRUE(vocabulary updateConstantFreedom cf)
    } else {
      Goal(facts, compoundFormulas,
           tasks ++ List(new UpdateConstantFreedomTask(cf, age)), age,
           eliminatedConstants, vocabulary, definedSyms, settings)
    }
  
  //////////////////////////////////////////////////////////////////////////////
  
  val subtrees : Seq[ProofTree] = List()
  
  override def toString : String = {
    "^ " + closingConstraint + "\n" +
    ". Facts: " + facts + "\n" +
    ". Arithmetic clauses: " + compoundFormulas.qfClauses + "\n" +
    ". Eagerly matched clauses: " + compoundFormulas.eagerQuantifiedClauses + "\n" +
    ". Lazily matched clauses: " + compoundFormulas.lazyQuantifiedClauses + "\n" +
    ". Tasks: " + tasks + "\n" +
    ". Eliminated constants: " + eliminatedConstants + "\n" +
    ". Free constants " + (if (fixedConstantFreedom) "(fixed)" else "(not fixed)") +
                     ": " + constantFreedom + "\n" +
    ". Defined constants: " + definedSyms
  }
}
